@@ -1,17 +1,39 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
-#include "mupdf/pdf.h"
+#include "pdf-annot-imp.h"
 
 #include <string.h>
 #include <assert.h>
 
 static void
-pdf_filter_xobject(fz_context *ctx, pdf_document *doc, pdf_obj *xobj, pdf_obj *page_res, pdf_filter_options *filter);
+pdf_filter_xobject(fz_context *ctx, pdf_document *doc, pdf_obj *xobj, pdf_obj *page_res, pdf_filter_options *filter, pdf_cycle_list *cycle_up);
 
 static void
-pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page_res, pdf_filter_options *filter);
+pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page_res, pdf_filter_options *filter, pdf_cycle_list *cycle_up);
 
 static void
-pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_obj *res, pdf_filter_options *filter)
+pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_obj *res, pdf_filter_options *filter, pdf_cycle_list *cycle_up)
 {
 	pdf_obj *obj;
 	int i, n;
@@ -33,7 +55,7 @@ pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_ob
 				if (g)
 				{
 					/* Transparency group XObject */
-					pdf_filter_xobject(ctx, doc, g, in_res, filter);
+					pdf_filter_xobject(ctx, doc, g, in_res, filter, cycle_up);
 				}
 			}
 		}
@@ -49,7 +71,7 @@ pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_ob
 			pdf_obj *pat = pdf_dict_get_val(ctx, obj, i);
 			if (pat && pdf_dict_get_int(ctx, pat, PDF_NAME(PatternType)) == 1)
 			{
-				pdf_filter_xobject(ctx, doc, pat, in_res, filter);
+				pdf_filter_xobject(ctx, doc, pat, in_res, filter, cycle_up);
 			}
 		}
 	}
@@ -66,7 +88,7 @@ pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_ob
 				pdf_obj *xobj = pdf_dict_get_val(ctx, obj, i);
 				if (xobj && pdf_dict_get(ctx, xobj, PDF_NAME(Subtype)) == PDF_NAME(Form))
 				{
-					pdf_filter_xobject(ctx, doc, xobj, in_res, filter);
+					pdf_filter_xobject(ctx, doc, xobj, in_res, filter, cycle_up);
 				}
 			}
 		}
@@ -82,7 +104,7 @@ pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_ob
 			pdf_obj *font = pdf_dict_get_val(ctx, obj, i);
 			if (font && pdf_dict_get(ctx, font, PDF_NAME(Subtype)) == PDF_NAME(Type3))
 			{
-				pdf_filter_type3(ctx, doc, font, in_res, filter);
+				pdf_filter_type3(ctx, doc, font, in_res, filter, cycle_up);
 			}
 		}
 	}
@@ -111,7 +133,8 @@ pdf_filter_content_stream(
 	pdf_filter_options *filter,
 	int struct_parents,
 	fz_buffer **out_buf,
-	pdf_obj **out_res)
+	pdf_obj **out_res,
+	pdf_cycle_list *cycle_up)
 {
 	pdf_processor *proc_buffer = NULL;
 	pdf_processor *proc_filter = NULL;
@@ -140,7 +163,7 @@ pdf_filter_content_stream(
 		}
 		pdf_close_processor(ctx, proc_buffer);
 
-		pdf_filter_resources(ctx, doc, in_res, *out_res, filter);
+		pdf_filter_resources(ctx, doc, in_res, *out_res, filter, cycle_up);
 	}
 	fz_always(ctx)
 	{
@@ -163,8 +186,9 @@ pdf_filter_content_stream(
 	shared between all off the CharProcs.
 */
 static void
-pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page_res, pdf_filter_options *filter)
+pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page_res, pdf_filter_options *filter, pdf_cycle_list *cycle_up)
 {
+	pdf_cycle_list cycle;
 	pdf_processor *proc_buffer = NULL;
 	pdf_processor *proc_filter = NULL;
 	pdf_obj *in_res;
@@ -185,7 +209,7 @@ pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page
 	assert(!filter->instance_forms);
 
 	/* Avoid recursive cycles! */
-	if (pdf_mark_obj(ctx, obj))
+	if (pdf_cycle(ctx, &cycle, cycle_up, obj))
 		return;
 
 	fz_try(ctx)
@@ -234,14 +258,13 @@ pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page
 			}
 		}
 
-		pdf_filter_resources(ctx, doc, in_res, out_res, filter);
+		pdf_filter_resources(ctx, doc, in_res, out_res, filter, &cycle);
 
 		if (filter->sanitize)
 			pdf_dict_put(ctx, obj, PDF_NAME(Resources), out_res);
 	}
 	fz_always(ctx)
 	{
-		pdf_unmark_obj(ctx, obj);
 		pdf_drop_obj(ctx, out_res);
 	}
 	fz_catch(ctx)
@@ -251,8 +274,9 @@ pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page
 }
 
 static void
-pdf_filter_xobject(fz_context *ctx, pdf_document *doc, pdf_obj *stm, pdf_obj *page_res, pdf_filter_options *filter)
+pdf_filter_xobject(fz_context *ctx, pdf_document *doc, pdf_obj *stm, pdf_obj *page_res, pdf_filter_options *filter, pdf_cycle_list *cycle_up)
 {
+	pdf_cycle_list cycle;
 	pdf_obj *struct_parents_obj;
 	int struct_parents;
 	pdf_obj *new_res = NULL;
@@ -276,17 +300,16 @@ pdf_filter_xobject(fz_context *ctx, pdf_document *doc, pdf_obj *stm, pdf_obj *pa
 	// TODO: don't clean objects more than once.
 
 	/* Avoid recursive cycles! */
-	if (pdf_mark_obj(ctx, stm))
+	if (pdf_cycle(ctx, &cycle, cycle_up, stm))
 		return;
 	fz_try(ctx)
 	{
-		pdf_filter_content_stream(ctx, doc, stm, old_res, fz_identity, filter, struct_parents, &new_buf, &new_res);
+		pdf_filter_content_stream(ctx, doc, stm, old_res, fz_identity, filter, struct_parents, &new_buf, &new_res, &cycle);
 		pdf_update_stream(ctx, doc, stm, new_buf, 0);
 		pdf_dict_put(ctx, stm, PDF_NAME(Resources), new_res);
 	}
 	fz_always(ctx)
 	{
-		pdf_unmark_obj(ctx, stm);
 		fz_drop_buffer(ctx, new_buf);
 		pdf_drop_obj(ctx, new_res);
 	}
@@ -295,14 +318,16 @@ pdf_filter_xobject(fz_context *ctx, pdf_document *doc, pdf_obj *stm, pdf_obj *pa
 }
 
 pdf_obj *
-pdf_filter_xobject_instance(fz_context *ctx, pdf_obj *old_xobj, pdf_obj *page_res, fz_matrix transform, pdf_filter_options *filter)
+pdf_filter_xobject_instance(fz_context *ctx, pdf_obj *old_xobj, pdf_obj *page_res, fz_matrix transform, pdf_filter_options *filter, pdf_cycle_list *cycle_up)
 {
+	pdf_cycle_list cycle;
 	pdf_document *doc = pdf_get_bound_document(ctx, old_xobj);
 	pdf_obj *new_xobj;
 	pdf_obj *new_res, *old_res;
 	fz_buffer *new_buf;
 	pdf_obj *struct_parents_obj;
 	int struct_parents;
+	fz_matrix matrix;
 
 	fz_var(new_xobj);
 	fz_var(new_buf);
@@ -320,19 +345,21 @@ pdf_filter_xobject_instance(fz_context *ctx, pdf_obj *old_xobj, pdf_obj *page_re
 	if (!old_res)
 		old_res = page_res;
 
-	if (pdf_mark_obj(ctx, old_xobj))
+	if (pdf_cycle(ctx, &cycle, cycle_up, old_xobj))
 		return pdf_keep_obj(ctx, old_xobj);
 
+	matrix = pdf_dict_get_matrix(ctx, old_xobj, PDF_NAME(Matrix));
+	transform = fz_concat(transform, matrix);
+
 	fz_try(ctx)
 	{
 		new_xobj = pdf_add_object_drop(ctx, doc, pdf_copy_dict(ctx, old_xobj));
-		pdf_filter_content_stream(ctx, doc, old_xobj, old_res, transform, filter, struct_parents, &new_buf, &new_res);
+		pdf_filter_content_stream(ctx, doc, old_xobj, old_res, transform, filter, struct_parents, &new_buf, &new_res, &cycle);
 		pdf_update_stream(ctx, doc, new_xobj, new_buf, 0);
 		pdf_dict_put(ctx, new_xobj, PDF_NAME(Resources), new_res);
 	}
 	fz_always(ctx)
 	{
-		pdf_unmark_obj(ctx, old_xobj);
 		fz_drop_buffer(ctx, new_buf);
 		pdf_drop_obj(ctx, new_res);
 	}
@@ -361,15 +388,16 @@ void pdf_filter_page_contents(fz_context *ctx, pdf_document *doc, pdf_page *page
 	contents = pdf_page_contents(ctx, page);
 	old_res = pdf_page_resources(ctx, page);
 
-	pdf_filter_content_stream(ctx, doc, contents, old_res, fz_identity, filter, struct_parents, &buffer, &new_res);
+	pdf_filter_content_stream(ctx, doc, contents, old_res, fz_identity, filter, struct_parents, &buffer, &new_res, NULL);
 
 	fz_try(ctx)
 	{
 		if (filter->end_page)
 			filter->end_page(ctx, buffer, filter->opaque);
-		if (pdf_is_array(ctx, contents))
+		/* If contents is not a stream it's an array of streams or missing. */
+		if (!pdf_is_stream(ctx, contents))
 		{
-			/* Create a new stream object to replace the array of streams. */
+			/* Create a new stream object to replace the array of streams or missing object. */
 			contents = pdf_add_object_drop(ctx, doc, pdf_new_dict(ctx, doc, 1));
 			pdf_dict_put_drop(ctx, page->obj, PDF_NAME(Contents), contents);
 		}
@@ -396,7 +424,7 @@ void pdf_filter_annot_contents(fz_context *ctx, pdf_document *doc, pdf_annot *an
 			pdf_obj *stm = pdf_dict_get_val(ctx, ap, i);
 			if (pdf_is_stream(ctx, stm))
 			{
-				pdf_filter_xobject(ctx, doc, stm, NULL, filter);
+				pdf_filter_xobject(ctx, doc, stm, NULL, filter, NULL);
 			}
 		}
 	}
@@ -494,22 +522,50 @@ pdf_redact_text_filter(fz_context *ctx, void *opaque, int *ucsbuf, int ucslen, f
 }
 
 static fz_pixmap *
-pdf_redact_image_imp(fz_context *ctx, fz_matrix ctm, fz_image *image, fz_pixmap *pixmap, fz_quad q)
+pdf_redact_image_imp(fz_context *ctx, fz_matrix ctm, fz_image *image, fz_pixmap *pixmap, fz_pixmap **pmask, fz_quad q)
 {
 	fz_matrix inv_ctm;
 	fz_irect r;
 	int x, y, k, n, bpp;
 	unsigned char white;
+	fz_pixmap *mask = *pmask;
+	int pixmap_cloned = 0;
 
 	if (!pixmap)
 	{
 		fz_pixmap *original = fz_get_pixmap_from_image(ctx, image, NULL, NULL, NULL, NULL);
+		int imagemask = image->imagemask;
+
 		fz_try(ctx)
+		{
 			pixmap = fz_clone_pixmap(ctx, original);
+			if (imagemask)
+				fz_invert_pixmap_alpha(ctx, pixmap);
+		}
 		fz_always(ctx)
 			fz_drop_pixmap(ctx, original);
 		fz_catch(ctx)
 			fz_rethrow(ctx);
+		pixmap_cloned = 1;
+	}
+
+	if (!mask && image->mask)
+	{
+		fz_pixmap *original = fz_get_pixmap_from_image(ctx, image->mask, NULL, NULL, NULL, NULL);
+
+		fz_try(ctx)
+		{
+			mask = fz_clone_pixmap(ctx, original);
+			*pmask = mask;
+		}
+		fz_always(ctx)
+		{
+			fz_drop_pixmap(ctx, original);
+			if (pixmap_cloned)
+				fz_drop_pixmap(ctx, pixmap);
+		}
+		fz_catch(ctx)
+			fz_rethrow(ctx);
 	}
 
 	n = pixmap->n - pixmap->alpha;
@@ -537,6 +593,21 @@ pdf_redact_image_imp(fz_context *ctx, fz_matrix ctm, fz_image *image, fz_pixmap
 		}
 	}
 
+	if (mask)
+	{
+		inv_ctm = fz_post_scale(fz_invert_matrix(ctm), mask->w, mask->h);
+		r = fz_round_rect(fz_transform_rect(fz_rect_from_quad(q), inv_ctm));
+		r.x0 = fz_clampi(r.x0, 0, mask->w);
+		r.x1 = fz_clampi(r.x1, 0, mask->w);
+		r.y1 = fz_clampi(mask->h - r.y1, 0, mask->h);
+		r.y0 = fz_clampi(mask->h - r.y0, 0, mask->h);
+		for (y = r.y1; y < r.y0; ++y)
+		{
+			unsigned char *s = &mask->samples[(size_t)y * mask->stride + (size_t)r.x0];
+			memset(s, 0xff, r.x1-r.x0);
+		}
+	}
+
 	return pixmap;
 }
 
@@ -588,6 +659,7 @@ static fz_image *
 pdf_redact_image_filter_pixels(fz_context *ctx, void *opaque, fz_matrix ctm, const char *name, fz_image *image)
 {
 	fz_pixmap *redacted = NULL;
+	fz_pixmap *mask = NULL;
 	pdf_page *page = opaque;
 	pdf_annot *annot;
 	pdf_obj *qp;
@@ -596,6 +668,7 @@ pdf_redact_image_filter_pixels(fz_context *ctx, void *opaque, fz_matrix ctm, con
 	int i, n;
 
 	fz_var(redacted);
+	fz_var(mask);
 
 	area = fz_transform_quad(fz_quad_from_rect(fz_unit_rect), ctm);
 
@@ -640,7 +713,7 @@ pdf_redact_image_filter_pixels(fz_context *ctx, void *opaque, fz_matrix ctm, con
 					{
 						q = pdf_to_quad(ctx, qp, i);
 						if (fz_is_quad_intersecting_quad(area, q))
-							redacted = pdf_redact_image_imp(ctx, ctm, image, redacted, q);
+							redacted = pdf_redact_image_imp(ctx, ctm, image, redacted, &mask, q);
 					}
 				}
 				else
@@ -648,7 +721,7 @@ pdf_redact_image_filter_pixels(fz_context *ctx, void *opaque, fz_matrix ctm, con
 					r = pdf_dict_get_rect(ctx, annot->obj, PDF_NAME(Rect));
 					q = fz_quad_from_rect(r);
 					if (fz_is_quad_intersecting_quad(area, q))
-						redacted = pdf_redact_image_imp(ctx, ctm, image, redacted, q);
+						redacted = pdf_redact_image_imp(ctx, ctm, image, redacted, &mask, q);
 				}
 			}
 		}
@@ -656,15 +729,36 @@ pdf_redact_image_filter_pixels(fz_context *ctx, void *opaque, fz_matrix ctm, con
 	fz_catch(ctx)
 	{
 		fz_drop_pixmap(ctx, redacted);
+		fz_drop_pixmap(ctx, mask);
 		fz_rethrow(ctx);
 	}
 
 	if (redacted)
 	{
+		int imagemask = image->imagemask;
+		fz_image *imask = fz_keep_image(ctx, image->mask);
+
+		fz_var(imask);
+
 		fz_try(ctx)
+		{
+			if (mask)
+			{
+				fz_drop_image(ctx, imask);
+				imask = NULL;
+				imask = fz_new_image_from_pixmap(ctx, mask, NULL);
+			}
 			image = fz_new_image_from_pixmap(ctx, redacted, NULL);
+			image->imagemask = imagemask;
+			image->mask = imask;
+			imask = NULL;
+		}
 		fz_always(ctx)
+		{
 			fz_drop_pixmap(ctx, redacted);
+			fz_drop_pixmap(ctx, mask);
+			fz_drop_image(ctx, imask);
+		}
 		fz_catch(ctx)
 			fz_rethrow(ctx);
 		return image;
@@ -673,7 +767,7 @@ pdf_redact_image_filter_pixels(fz_context *ctx, void *opaque, fz_matrix ctm, con
 }
 
 static int
-pdf_redact_page_link(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_rect area)
+rect_touches_redactions(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_rect area)
 {
 	pdf_annot *annot;
 	pdf_obj *qp;
@@ -726,7 +820,7 @@ pdf_redact_page_links(fz_context *ctx, pdf_document *doc, pdf_page *page)
 		if (pdf_dict_get(ctx, link, PDF_NAME(Subtype)) == PDF_NAME(Link))
 		{
 			area = pdf_dict_get_rect(ctx, link, PDF_NAME(Rect));
-			if (pdf_redact_page_link(ctx, doc, page, area))
+			if (rect_touches_redactions(ctx, doc, page, area))
 			{
 				pdf_array_delete(ctx, annots, k);
 				continue;
@@ -736,6 +830,27 @@ pdf_redact_page_links(fz_context *ctx, pdf_document *doc, pdf_page *page)
 	}
 }
 
+static void
+pdf_redact_page_annotations(fz_context *ctx, pdf_document *doc, pdf_page *page)
+{
+	pdf_annot *annot;
+	fz_rect area;
+
+restart:
+	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
+	{
+		if (pdf_annot_type(ctx, annot) == PDF_ANNOT_FREE_TEXT)
+		{
+			area = pdf_dict_get_rect(ctx, pdf_annot_obj(ctx, annot), PDF_NAME(Rect));
+			if (rect_touches_redactions(ctx, doc, page, area))
+			{
+				pdf_delete_annot(ctx, page, annot);
+				goto restart;
+			}
+		}
+	}
+}
+
 int
 pdf_redact_page(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_redact_options *opts)
 {
@@ -773,24 +888,33 @@ pdf_redact_page(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_redact_o
 	if (!has_redactions)
 		return 0;
 
-	pdf_filter_page_contents(ctx, doc, page, &filter);
-	pdf_redact_page_links(ctx, doc, page);
-
-	annot = pdf_first_annot(ctx, page);
-	while (annot)
+	pdf_begin_operation(ctx, doc, "Redact page");
+	fz_try(ctx)
 	{
-		if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
-		{
-			pdf_delete_annot(ctx, page, annot);
-			annot = pdf_first_annot(ctx, page);
-		}
-		else
-		{
-			annot = pdf_next_annot(ctx, annot);
-		}
-	}
+		pdf_filter_page_contents(ctx, doc, page, &filter);
+		pdf_redact_page_links(ctx, doc, page);
+		pdf_redact_page_annotations(ctx, doc, page);
 
-	doc->redacted = 1;
+		annot = pdf_first_annot(ctx, page);
+		while (annot)
+		{
+			if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
+			{
+				pdf_delete_annot(ctx, page, annot);
+				annot = pdf_first_annot(ctx, page);
+			}
+			else
+			{
+				annot = pdf_next_annot(ctx, annot);
+			}
+		}
+
+		doc->redacted = 1;
+	}
+	fz_always(ctx)
+		pdf_end_operation(ctx, doc);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 
 	return 1;
 }
