@@ -103,7 +103,7 @@ static void Ap_join(js_State *J)
 		seplen = 1;
 	}
 
-	if (len == 0) {
+	if (len <= 0) {
 		js_pushliteral(J, "");
 		return;
 	}
@@ -123,11 +123,15 @@ static void Ap_join(js_State *J)
 		n += strlen(r);
 
 		if (k == 0) {
-			out = js_malloc(J, n);
+			if (n > JS_STRLIMIT)
+				js_rangeerror(J, "invalid string length");
+			out = js_malloc(J, (int)n);
 			strcpy(out, r);
 		} else {
 			n += seplen;
-			out = js_realloc(J, out, n);
+			if (n > JS_STRLIMIT)
+				js_rangeerror(J, "invalid string length");
+			out = js_realloc(J, out, (int)n);
 			strcat(out, sep);
 			strcat(out, r);
 		}
@@ -248,72 +252,98 @@ static void Ap_slice(js_State *J)
 			js_setindex(J, -2, n);
 }
 
-static int compare(js_State *J, int x, int y, int *hasx, int *hasy, int hasfn)
+struct sortslot {
+	js_Value v;
+	js_State *J;
+};
+
+static int sortcmp(const void *avoid, const void *bvoid)
 {
+	const struct sortslot *aslot = avoid, *bslot = bvoid;
+	const js_Value *a = &aslot->v, *b = &bslot->v;
+	js_State *J = aslot->J;
 	const char *sx, *sy;
+	double v;
 	int c;
 
-	*hasx = js_hasindex(J, 0, x);
-	*hasy = js_hasindex(J, 0, y);
-
-	if (*hasx && *hasy) {
-		int unx = js_isundefined(J, -2);
-		int uny = js_isundefined(J, -1);
-		if (unx && uny) return 0;
-		if (unx) return 1;
-		if (uny) return -1;
-
-		if (hasfn) {
-			js_copy(J, 1); /* copy function */
-			js_pushundefined(J); /* set this object */
-			js_copy(J, -4); /* copy x */
-			js_copy(J, -4); /* copy y */
-			js_call(J, 2);
-			c = js_tonumber(J, -1);
-			js_pop(J, 1);
-			return c;
-		}
+	int unx = (a->type == JS_TUNDEFINED);
+	int uny = (b->type == JS_TUNDEFINED);
+	if (unx) return !uny;
+	if (uny) return -1;
 
+	if (js_iscallable(J, 1)) {
+		js_copy(J, 1); /* copy function */
+		js_pushundefined(J);
+		js_pushvalue(J, *a);
+		js_pushvalue(J, *b);
+		js_call(J, 2);
+		v = js_tonumber(J, -1);
+		c = (v == 0) ? 0 : (v < 0) ? -1 : 1;
+		js_pop(J, 1);
+	} else {
+		js_pushvalue(J, *a);
+		js_pushvalue(J, *b);
 		sx = js_tostring(J, -2);
 		sy = js_tostring(J, -1);
-		return strcmp(sx, sy);
+		c = strcmp(sx, sy);
+		js_pop(J, 2);
 	}
-
-	if (*hasx) return -1;
-	if (*hasy) return 1;
-	return 0;
+	return c;
 }
 
 static void Ap_sort(js_State *J)
 {
-	int len, i, k;
-	int hasx, hasy, hasfn;
+	struct sortslot * volatile array = NULL;
+	int i, n, len;
 
 	len = js_getlength(J, 0);
-
-	hasfn = js_iscallable(J, 1);
-	hasx = hasy = 0;
-
-	for (i = 1; i < len; ++i) {
-		k = i;
-		while (k > 0 && compare(J, k - 1, k, &hasx, &hasy, hasfn) > 0) {
-			if (hasx && hasy) {
-				js_setindex(J, 0, k - 1);
-				js_setindex(J, 0, k);
-			} else if (hasx) {
-				js_delindex(J, 0, k - 1);
-				js_setindex(J, 0, k);
-			} else if (hasy) {
-				js_setindex(J, 0, k - 1);
-				js_delindex(J, 0, k);
-			}
-			hasx = hasy = 0;
-			--k;
-		}
-		if (hasx + hasy > 0)
-			js_pop(J, hasx + hasy);
+	if (len <= 0) {
+		js_copy(J, 0);
+		return;
 	}
 
+	if (len >= INT_MAX / (int)sizeof(*array))
+		js_rangeerror(J, "array is too large to sort");
+
+	/* Holding objects where the GC cannot see them is illegal, but if we
+	 * don't allow the GC to run we can use qsort() on a temporary array of
+	 * js_Values for fast sorting.
+	 */
+	++J->gcpause;
+
+	if (js_try(J)) {
+		--J->gcpause;
+		js_free(J, array);
+		js_throw(J);
+	}
+
+	array = js_malloc(J, len * sizeof *array);
+
+	n = 0;
+	for (i = 0; i < len; ++i) {
+		if (js_hasindex(J, 0, i)) {
+			array[n].v = *js_tovalue(J, -1);
+			array[n].J = J;
+			js_pop(J, 1);
+			++n;
+		}
+	}
+
+	qsort(array, n, sizeof *array, sortcmp);
+
+	for (i = 0; i < n; ++i) {
+		js_pushvalue(J, array[i].v);
+		js_setindex(J, 0, i);
+	}
+	for (i = n; i < len; ++i) {
+		js_delindex(J, 0, i);
+	}
+
+	--J->gcpause;
+
+	js_endtry(J);
+	js_free(J, array);
+
 	js_copy(J, 0);
 }
 
@@ -397,9 +427,20 @@ static void Ap_unshift(js_State *J)
 
 static void Ap_toString(js_State *J)
 {
-	int top = js_gettop(J);
-	js_pop(J, top - 1);
-	Ap_join(J);
+	if (!js_iscoercible(J, 0))
+		js_typeerror(J, "'this' is not an object");
+	js_getproperty(J, 0, "join");
+	if (!js_iscallable(J, -1)) {
+		js_pop(J, 1);
+		// TODO: call Object.prototype.toString implementation directly
+		js_getglobal(J, "Object");
+		js_getproperty(J, -1, "prototype");
+		js_rot2pop1(J);
+		js_getproperty(J, -1, "toString");
+		js_rot2pop1(J);
+	}
+	js_copy(J, 0);
+	js_call(J, 0);
 }
 
 static void Ap_indexOf(js_State *J)
